@@ -1,6 +1,6 @@
 from logging import Logger
 from pathlib import Path
-from typing import Any, Dict, List
+from typing import TYPE_CHECKING, Any, Dict, List, Tuple
 
 import numpy as np
 import pandas as pd
@@ -10,6 +10,9 @@ from snpio.utils import misc, sequence_tools
 from snpio.utils.logging import LoggerManager
 from snpio.utils.misc import IUPAC
 
+if TYPE_CHECKING:
+    from snpio.read_input.genotype_data import GenotypeData
+
 
 class GenotypeEncoder:
     """Encode genotypes to various formats suitable for machine learning.
@@ -45,7 +48,7 @@ class GenotypeEncoder:
         replace_vals (List[str]): List of values to replace missing values with.
     """
 
-    def __init__(self, genotype_data: Any) -> None:
+    def __init__(self, genotype_data: "GenotypeData") -> None:
         """Initialize the GenotypeEncoder object.
 
         This class provides methods to encode genotypes to various formats suitable for machine learning, including 012, one-hot, and integer encodings, as well as the inverse operations.
@@ -75,168 +78,192 @@ class GenotypeEncoder:
 
         self.iupac = IUPAC(logger=self.logger)
 
-    def read_012(self) -> None:
-        """Read 012-encoded comma-delimited file.
-
-        This method reads a 012-encoded comma-delimited file and stores the data in the GenotypeEncoder object.
-
-        Raises:
-            ValueError: Sequences differ in length.
-        """
-        self.logger.info(f"Reading 012-encoded file: {self.filename}...")
-
-        self._check_filetype("encoded")
-        snp_data, num_snps = [], []
-
-        with open(self.filename, "r") as fin:
-            num_inds = 0
-            for line in fin:
-                line = line.strip()
-                if not line:
-                    continue
-                cols = line.split(",")
-                inds = cols[0]
-                snps = cols[1:]
-                num_snps.append(len(snps))
-                num_inds += 1
-                snp_data.append(snps)
-                self.samples.append(inds)
-
-        if len(list(set(num_snps))) > 1:
-            msg = "All sequences must be the same length. At least one sequence differs in length from the others"
-            self.logger.error(msg)
-            raise exceptions.SequenceLengthError(msg)
-
-        miss_vals = ["NA", np.nan, pd.NA, "N", ".", "-", "?"]
-        df = pd.DataFrame(snp_data)
-        df = df.replace(miss_vals, ["-9"] * len(miss_vals)).astype(int)
-
-        # Decodes 012 and converts to self.snp_data
-        self.genotypes_012 = df
-
-        self.ref = None
-        self.alt = None
-
-        self.logger.info("012 file successfully loaded!")
-        self.logger.info(
-            f"\nFound:\n"
-            f"\t{self.num_inds} individuals...\n"
-            f"\t{self.num_snps} SNPs...\n"
-        )
-
     def convert_012(self, snps: List[List[str]]) -> List[List[int]]:
-        """Encode IUPAC nucleotides as 0 (reference), 1 (heterozygous), and 2 (alternate) alleles.
+        """Convert IUPAC/diploid genotype strings to 012 encoding (0=REF, 1=HET, 2=ALT, -9=missing).
 
-        This method encodes IUPAC nucleotides as 0 (reference), 1 (heterozygous), and 2 (alternate) alleles.
+        This encoder is robust to: Diploids written as "A/G" or "A|G" IUPAC ambiguity codes (e.g., R= A/G). Multi-ALT columns (forces biallelic by collapsing all ALTs to one ALT class). Common missing tokens ("N", "-", ".", "?", "./.", "-9"). It also logs monomorphic, non-biallelic, and all-missing columns (and writes indices).
 
         Args:
-            snps (List[List[str]]): 2D list of genotypes of shape (n_samples, n_sites).
+            snps (List[List[str]]): Genotypes as 2D list of shape (n_samples, n_sites). Elements may be single bases ("A") or diploids ("A/G", "A|G"), or IUPAC single-letter codes.
 
         Returns:
-            List[List[int]]: Encoded 012 genotypes.
+            List[List[int]]: 012-encoded genotypes with missing as -9.
 
-        Warning:
-            Monomorphic sites are detected and encoded as 0 (reference).
-
-            Non-biallelic sites are detected and forced to be bi-allelic.
-
-            Sites with all missing data are detected and excluded from the alignment.
+        Notes:
+            - If a column is monomorphic, everything non-missing becomes 0 (REF); heterozygous forms or non-REF tokens map to 1 only when necessary during forced decisions.
+            - If a column has >2 alleles, all non-REF non-missing are collapsed to ALT (=2).
+            - Membership checks are performed against an explicit set of ALT alleles (no iterator pitfalls).
         """
-        snps_012 = []
-        new_snps = []
-        monomorphic_sites = []
-        non_biallelic_sites = []
-        all_missing = []
+        # ---- Local helpers  ----
+        IUPAC_TO_ALLELES = {
+            "A": {"A"},
+            "C": {"C"},
+            "G": {"G"},
+            "T": {"T"},
+            "R": {"A", "G"},
+            "Y": {"C", "T"},
+            "S": {"G", "C"},
+            "W": {"A", "T"},
+            "K": {"G", "T"},
+            "M": {"A", "C"},
+            "N": set(),
+            "-": set(),
+            ".": set(),
+            "?": set(),
+        }
 
-        for i in range(0, len(snps)):
-            new_snps.append([])
+        MISSING = {"-9", "N", "-", ".", "?", "./."}
 
-        for j in range(0, len(snps[0])):
-            loc = []
-            for i in range(0, len(snps)):
-                loc.append(snps[i][j].upper())
+        def _split_diploid(tok: str) -> list[str]:
+            t = tok.strip()
+            if "/" in t:
+                return t.split("/")
+            if "|" in t:
+                return t.split("|")
+            if len(t) == 1 and t.upper() in IUPAC_TO_ALLELES:
+                return list(IUPAC_TO_ALLELES[t.upper()])
+            return [t]  # fallback
 
-            if all(x == "N" for x in loc):
+        def _flatten_alts(mc) -> tuple[str, set[str]]:
+            # mc expected like ("A","G") or ("A", ["C","G"])
+            if not mc:
+                return "N", set()
+            ref = str(mc[0])
+            alts = set()
+            for a in mc[1:]:
+                if a is None:
+                    continue
+                if isinstance(a, (list, tuple, set)):
+                    for x in a:
+                        alts.add(str(x))
+                else:
+                    alts.add(str(a))
+            return ref, alts
+
+        def _encode_token(g: str, ref: str, alts: set[str]) -> int:
+            if g in MISSING or g.strip() == "":
+                return -9
+            alleles = _split_diploid(g)
+            if not alleles:  # from IUPAC N/ambiguous empty
+                return -9
+
+            ref_u = ref.upper()
+            alts_u = {a.upper() for a in alts} if alts else set()
+            a_u = [a.upper() for a in alleles]
+
+            # Homozygous REF
+            if len(a_u) == 1 and a_u[0] == ref_u:
+                return 0
+            if len(a_u) == 2 and a_u[0] == ref_u and a_u[1] == ref_u:
+                return 0
+
+            # Homozygous ALT (same ALT allele twice)
+            if len(a_u) == 1 and a_u[0] in alts_u:
+                return 2
+            if (
+                len(a_u) == 2
+                and a_u[0] in alts_u
+                and a_u[1] in alts_u
+                and a_u[0] == a_u[1]
+            ):
+                return 2
+
+            # Mixed ref/alt → heterozygote
+            has_ref = any(a == ref_u for a in a_u)
+            has_alt = any(a in alts_u for a in a_u)
+            if has_ref and has_alt:
+                return 1
+
+            # Ambiguity spanning multiple bases (e.g., IUPAC) → heterozygote
+            if len(set(a_u)) > 1:
+                return 1
+
+            # Unrecognized but non-REF → treat as heterozygote conservatively
+            return 1
+
+        # ---- Body ----
+        n_samples = len(snps)
+        n_sites = 0 if n_samples == 0 else len(snps[0])
+
+        new_snps: list[list[int]] = [[] for _ in range(n_samples)]
+        monomorphic_sites: list[int] = []
+        non_biallelic_sites: list[int] = []
+        all_missing: list[int] = []
+
+        for j in range(n_sites):
+            # column j
+            loc = [str(snps[i][j]).upper() for i in range(n_samples)]
+
+            # All missing?
+            if all(x in {"N", "-", ".", "?", "./.", "-9"} for x in loc):
                 all_missing.append(j)
+                # still need to append -9 to keep rectangular shape
+                for i in range(n_samples):
+                    new_snps[i].append(-9)
                 continue
+
+            # Count unique biological alleles ignoring missing
             num_alleles = sequence_tools.count_alleles(loc)
-            if num_alleles != 2:
-                # If monomorphic
-                if num_alleles < 2:
-                    monomorphic_sites.append(j)
-                    try:
-                        ref = list(
-                            map(
-                                sequence_tools.get_major_allele,
-                                loc,
-                                [False for x in loc],
-                            )
-                        )
-                        ref = str(ref[0])
-                    except IndexError:
-                        ref = list(
-                            map(
-                                sequence_tools.get_major_allele,
-                                loc,
-                                [False for x in loc],
-                            )
-                        )
-                    alt = None
 
-                    for i in range(0, len(snps)):
-                        if loc[i] in ["-", "-9", "N"]:
-                            new_snps[i].append(-9)
-
-                        elif loc[i] == ref:
-                            new_snps[i].append(0)
-
-                        else:
-                            new_snps[i].append(1)
-
-                # If >2 alleles
-                elif num_alleles > 2:
-                    non_biallelic_sites.append(j)
-                    all_alleles = sequence_tools.get_major_allele(loc)
-                    all_alleles = [str(x[0]) for x in all_alleles]
-                    ref = all_alleles.pop(0)
-                    alt = all_alleles.pop(0)
-                    others = all_alleles
-
-                    for i in range(0, len(snps)):
-                        if loc[i] in ["-", "-9", "N"]:
-                            new_snps[i].append(-9)
-
-                        elif loc[i] == ref:
-                            new_snps[i].append(0)
-
-                        elif loc[i] == alt:
-                            new_snps[i].append(2)
-
-                        # Force biallelic
-                        elif loc[i] in others:
-                            new_snps[i].append(2)
-
-                        else:
-                            new_snps[i].append(1)
-
-            else:
-                ref, alt = sequence_tools.get_major_allele(loc)
-                ref = str(ref)
-                alt = str(alt)
-
-                for i in range(0, len(snps)):
-                    if loc[i] in ["-", "-9", "N"]:
+            if num_alleles < 2:
+                # Monomorphic (or effectively so)
+                monomorphic_sites.append(j)
+                mc = sequence_tools.get_major_allele(
+                    loc, vcf=self.genotype_data.from_vcf
+                )
+                ref = "N" if not mc else str(mc[0])
+                for i in range(n_samples):
+                    g = loc[i]
+                    if g in {"-", "-9", "N", ".", "?", "./."}:
                         new_snps[i].append(-9)
-
-                    elif loc[i] == ref:
+                    elif g == ref or _encode_token(g, ref, set()) == 0:
                         new_snps[i].append(0)
-
-                    elif loc[i] == alt:
-                        new_snps[i].append(2)
-
                     else:
+                        # Any non-missing non-REF gets 1 (defensive)
                         new_snps[i].append(1)
 
+            elif num_alleles > 2:
+                # Force biallelic: collapse all non-REF to ALT
+                non_biallelic_sites.append(j)
+                all_alleles = sequence_tools.get_major_allele(
+                    loc, vcf=self.genotype_data.from_vcf
+                )
+                # make a flat ordered list like [ref, alt1, alt2, ...]
+                flat = []
+                for a in all_alleles:
+                    if isinstance(a, (list, tuple)):
+                        flat.extend([str(x) for x in a])
+                    else:
+                        flat.append(str(a))
+                if not flat:
+                    ref = "N"
+                    alts = set()
+                else:
+                    ref = flat[0]
+                    alts = set(flat[1:])  # everything else is ALT class
+                for i in range(n_samples):
+                    g = loc[i]
+                    if g in {"-", "-9", "N", ".", "?", "./."}:
+                        new_snps[i].append(-9)
+                    else:
+                        code = _encode_token(g, ref, alts)
+                        # ensure collapse: any non-REF non-missing that isn't het becomes ALT
+                        if code == 1:
+                            new_snps[i].append(1)
+                        elif code == 0:
+                            new_snps[i].append(0)
+                        else:
+                            new_snps[i].append(2)
+            else:
+                # Properly biallelic
+                mc = sequence_tools.get_major_allele(
+                    loc, vcf=self.genotype_data.from_vcf
+                )
+                ref, alts = _flatten_alts(mc)
+                for i in range(n_samples):
+                    new_snps[i].append(_encode_token(loc[i], ref, alts))
+
+        # ---- Logging side-effects preserved ----
         if self.genotype_data.was_filtered:
             outdir = Path(f"{self.prefix}_output", "nremover", "logs")
         else:
@@ -244,39 +271,31 @@ class GenotypeEncoder:
         outdir.mkdir(exist_ok=True, parents=True)
 
         if monomorphic_sites:
-            # TODO: Check here if column is all missing.
-            # TODO: What to do in this case? Error out?
-            outfile = outdir / "monomorphic_sites_mqc.txt"
-            with open(outfile, "w") as fout:
-                mono_sites = [str(x) for x in monomorphic_sites]
-                fout.write(",".join(mono_sites))
-
-            self.logger.warning(
-                f"Monomorphic sites detected. You can check the monomorphic locus indices in the following log file: {outfile}"
+            (outdir / "monomorphic_sites_mqc.txt").write_text(
+                ",".join(map(str, monomorphic_sites))
+            )
+            self.logger.info(
+                f"Monomorphic sites detected; indices written to: {outdir/'monomorphic_sites_mqc.txt'}"
             )
 
         if non_biallelic_sites:
-            outfile = outdir / "non_biallelic_sites_mqc.txt"
-            with open(outfile, "w") as fout:
-                nba = [str(x) for x in non_biallelic_sites]
-                fout.write(",".join(nba))
-
-            self.logger.warning(
-                f"SNP column indices listed in the log file {outfile} had >2 alleles and was forced to be bi-allelic. If that is not desired, please fix or remove the column and re-run."
+            (outdir / "non_biallelic_sites_mqc.txt").write_text(
+                ",".join(map(str, non_biallelic_sites))
+            )
+            self.logger.info(
+                f">2-allele columns collapsed to biallelic; indices written to: {outdir/'non_biallelic_sites_mqc.txt'}"
             )
 
-        if all_missing:
+        if any(idx in all_missing for idx in range(n_sites)):
             self.genotype_data.all_missing_idx = all_missing
-            outfile = outdir / "all_missing_sites_mqc.txt"
-            with open(outfile, "w") as fout:
-                fout.write(",".join([str(x) for x in all_missing]))
-
+            (outdir / "all_missing_sites_mqc.txt").write_text(
+                ",".join(map(str, all_missing))
+            )
             self.logger.warning(
-                f"SNP column indices found in the log file {outfile} had all missing data and were excluded from the alignment."
+                f"All-missing columns excluded; indices written to: {outdir/'all_missing_sites_mqc.txt'}"
             )
 
-        snps_012 = [s for s in new_snps]
-        return snps_012
+        return [row for row in new_snps]
 
     def convert_onehot(
         self,
@@ -466,190 +485,194 @@ class GenotypeEncoder:
         return np.array(decoded_outer_list)
 
     def decode_012(
-        self,
-        X: np.ndarray | pd.DataFrame | List[List[int]],
-        write_output: bool = True,
-        is_nuc: bool = False,
-    ):
-        """Decode 012-encoded or 0-9 integer-encoded imputed data to STRUCTURE or PHYLIP format.
+        self, X: np.ndarray | pd.DataFrame | List[List[int]], is_nuc: bool = False
+    ) -> np.ndarray:
+        """Decode 012 or 0-9 integer encodings to single-character IUPAC nucleotides.
 
-        This method decodes 012-encoded or 0-9 integer-encoded imputed data to IUPAC format. The decoded data can be saved to a file or returned as a DataFrame.
+        Always returns single-character IUPAC codes: A, C, G, T for homozygotes; R, Y, S, W, K, M for heterozygotes; N for missing.
+
+        Modes:
+        1) Standard 012 (0=REF, 1=HET, 2=ALT, -9=missing) using per-locus REF/ALT from GenotypeData.
+        2) IUPAC-integer mode (``is_nuc=True``) using SNPio's updated order:
+            A=0, C=1, G=2, T=3, W=4, R=5, M=6, K=7, Y=8, S=9, N=-9.
 
         Args:
-            X (pandas.DataFrame | numpy.ndarray | List[List[int]]): Imputed data to decode, encoded as 012 or 0-9 integers.
-
-            write_output (bool): If True, save the decoded output to a file. If False, return the decoded data as a DataFrame. Defaults to True.
-
-            is_nuc (bool): Whether the encoding is based on nucleotides instead of 012. Defaults to False.
+            X: Matrix of 012 or 0-9 IUPAC integers.
+            is_nuc: If True, interpret inputs as 0-9 IUPAC integers (A=0, C=1, G=2, T=3, ...).
 
         Returns:
-            str | pandas.DataFrame: If write_output is True, returns the filename where the imputed data was written. If write_output is False, returns the decoded data as a DataFrame.
+            np.ndarray: Same shape as X, dtype '<U1', with single-character IUPAC codes.
+
+        Raises:
+            ValueError: If REF/ALT metadata are unavailable for 012 decoding.
         """
         df = misc.validate_input_type(X, return_type="df")
 
-        nuc = {
-            "A/A": "A",
-            "T/T": "T",
-            "G/G": "G",
-            "C/C": "C",
-            "A/G": "R",
-            "G/A": "R",
-            "C/T": "Y",
-            "T/C": "Y",
-            "G/C": "S",
-            "C/G": "S",
-            "A/T": "W",
-            "T/A": "W",
-            "G/T": "K",
-            "T/G": "K",
-            "A/C": "M",
-            "C/A": "M",
-            "N/N": "N",
+        # IUPAC ambiguity mapping (unordered pairs → code) for 012→IUPAC.
+        pair_to_iupac = {
+            frozenset(("A", "G")): "R",
+            frozenset(("C", "T")): "Y",
+            frozenset(("G", "C")): "S",
+            frozenset(("A", "T")): "W",
+            frozenset(("G", "T")): "K",
+            frozenset(("A", "C")): "M",
         }
-
-        ft = self.filetype.lower()
-
-        is_phylip = False
-        if ft == "phylip" or ft == "vcf":
-            is_phylip = True
-
-        df_decoded = df.copy()
-        df_decoded = df.copy().astype(object)
+        valid_bases = {"A", "C", "G", "T"}
 
         if is_nuc:
-            classes_int = range(10)
-            classes_string = [str(x) for x in classes_int]
-            if is_phylip:
-                gt = ["A", "T", "G", "C", "W", "R", "M", "K", "Y", "S", "N"]
-            else:
-                gt = [
-                    "1/1",
-                    "2/2",
-                    "3/3",
-                    "4/4",
-                    "1/2",
-                    "1/3",
-                    "1/4",
-                    "2/3",
-                    "2/4",
-                    "3/4",
-                    "-9/-9",
-                ]
-            d = dict(zip(classes_int, gt))
-            dstr = dict(zip(classes_string, gt))
-            d.update(dstr)
-            dreplace = {col: d for col in list(df.columns)}
+            # UPDATED SNPio order: A=0, C=1, G=2, T=3, W=4, R=5, M=6, K=7, Y=8,
+            # S=9, N=-9
+            iupac_list = ["A", "C", "G", "T", "W", "R", "M", "K", "Y", "S"]
+            mapping = {i: iupac_list[i] for i in range(10)}
+            mapping[-9] = "N"
+            mapping[-1] = "N"
 
-        else:
-            dreplace = dict()
-            for col, ref, alt in zip(df.columns, self._ref, self._alt):
-                # if site is monomorphic, set alt and ref state the
-                # same
-                if alt is None:
-                    alt = ref
-                ref2 = f"{ref}/{ref}"
-                alt2 = f"{alt}/{alt}"
-                het2 = f"{ref}/{alt}"
+            # accept strings too
+            mapping.update({str(k): v for k, v in mapping.items()})
+            return df.replace(mapping).to_numpy(dtype="<U1")
 
-                if is_phylip:
-                    ref2 = nuc[ref2]
-                    alt2 = nuc[alt2]
-                    het2 = nuc[het2]
+        # ---- Standard 012 decoding using REF/ALT per column ----
+        ref_alleles = getattr(self.genotype_data, "ref", None)
+        alt_alleles = getattr(self.genotype_data, "alt", None)
 
-                d = {
-                    "0": ref2,
-                    0: ref2,
-                    "1": het2,
-                    1: het2,
-                    "2": alt2,
-                    2: alt2,
-                    "-9": "N",
+        if ref_alleles is None or alt_alleles is None:
+            msg = "Reference and alternate alleles are not available in GenotypeData; cannot decode 012 matrix."
+            self.logger.error(msg)
+            raise ValueError(msg)
+
+        df_out = df.copy().astype(object)
+        for j, col in enumerate(df_out.columns):
+            ref = ref_alleles[j]
+            alt = alt_alleles[j]
+
+            if isinstance(alt, (list, tuple)):
+                alt = alt[0] if len(alt) > 0 else None
+
+            ref = "" if ref is None else str(ref).upper()
+            alt = ref if alt is None else str(alt).upper()
+
+            ref_is_std = ref in valid_bases
+            alt_is_std = alt in valid_bases
+
+            if ref_is_std and alt_is_std:
+                het_code = (
+                    ref if ref == alt else pair_to_iupac.get(frozenset((ref, alt)), "N")
+                )
+                col_map = {
+                    0: ref,
+                    "0": ref,
+                    1: het_code,
+                    "1": het_code,
+                    2: alt,
+                    "2": alt,
                     -9: "N",
+                    "-9": "N",
+                    -1: "N",
+                    "-1": "N",
+                }
+            elif ref_is_std and not alt_is_std:
+                col_map = {
+                    0: ref,
+                    "0": ref,
+                    1: "N",
+                    "1": "N",
+                    2: "N",
+                    "2": "N",
+                    -9: "N",
+                    "-9": "N",
+                    -1: "N",
+                    "-1": "N",
+                }
+            elif not ref_is_std and alt_is_std:
+                col_map = {
+                    0: "N",
+                    "0": "N",
+                    1: "N",
+                    "1": "N",
+                    2: alt,
+                    "2": alt,
+                    -9: "N",
+                    "-9": "N",
+                    -1: "N",
+                    "-1": "N",
                 }
-                dreplace[col] = d
-
-        df_decoded = df_decoded.replace(dreplace)
-
-        if write_output:
-            if self.genotype_data.was_filtered:
-                outdir = Path(f"{self.prefix}_output", "nremover", "alignments")
             else:
-                outdir = Path(f"{self.prefix}_output", "alignments")
-            outdir.mkdir(exist_ok=True, parents=True)
+                col_map = {
+                    0: "N",
+                    "0": "N",
+                    1: "N",
+                    "1": "N",
+                    2: "N",
+                    "2": "N",
+                    -9: "N",
+                    "-9": "N",
+                    -1: "N",
+                    "-1": "N",
+                }
 
-        if ft.startswith("structure"):
-            if ft.startswith("structure2row"):
-                for col in df_decoded.columns:
-                    df_decoded[col] = (
-                        df_decoded[col]
-                        .str.split("/")
-                        .apply(lambda x: list(map(int, x)))
-                    )
+            df_out[col] = df_out[col].map(col_map)
 
-                df_decoded.insert(0, "sampleID", self._samples)
-                df_decoded.insert(1, "popID", self._populations)
+        return df_out.to_numpy(dtype="<U1")
 
-                # Transform each element to a separate row.
-                df_decoded = (
-                    df_decoded.set_index(["sampleID", "popID"])
-                    .apply(pd.Series.explode)
-                    .reset_index()
-                )
+    def encode_alleles_two_channel(
+        self, snp_data: np.ndarray
+    ) -> Tuple[np.ndarray, np.ndarray]:
+        """Convert IUPAC genotypes to two integer allele matrices.
 
-            elif ft.startswith("structure1row"):
-                df_decoded = pd.concat(
-                    [
-                        df_decoded[c]
-                        .astype(str)
-                        .str.split("/", expand=True)
-                        .add_prefix(f"{c}_")
-                        for c in df_decoded.columns
-                    ],
-                    axis=1,
-                )
+        This method encodes the SNP data into two separate matrices representing the two alleles for each sample and locus. Each matrix will have shape (N_samples, N_loci), where each entry is an integer representing one of the two alleles (reference or alternate).
 
-            elif ft == "structure":
-                for col in df_decoded.columns:
-                    df_decoded[col] = (
-                        df_decoded[col]
-                        .str.split("/")
-                        .apply(lambda x: list(map(int, x)))
-                    )
+        Args:
+            snp_data (np.ndarray): An (n_samples x n_loci) numpy array of IUPAC-encoded genotypes, where each entry is a single character string representing the genotype (e.g., "A", "C", "G", "T", "N", "-", etc.). Heterozygous genotypes are represented by ambiguity codes (e.g., "W", "S", "M", "K", "R", "Y").
 
-                df_decoded.insert(0, "sampleID", self._samples)
-                df_decoded.insert(1, "popID", self._populations)
+        Returns:
+            Tuple[np.ndarray, np.ndarray]: Two matrices where each row corresponds to a sample and each column to a locus. The first matrix contains the first allele (allele1) and the second matrix contains the second allele (allele2).
+        """
+        IUPAC_MAP = self.iupac.get_two_channel_iupac()
+        n_samples, n_loci = snp_data.shape
+        allele1 = np.full((n_samples, n_loci), -1, dtype=np.int8)
+        allele2 = np.full((n_samples, n_loci), -1, dtype=np.int8)
+        for i in range(n_samples):
+            for j in range(n_loci):
+                a1, a2 = IUPAC_MAP.get(snp_data[i, j].upper(), (-1, -1))
+                allele1[i, j], allele2[i, j] = a1, a2
+        return allele1, allele2
 
-                # Transform each element to a separate row.
-                df_decoded = (
-                    df_decoded.set_index(["sampleID", "popID"])
-                    .apply(pd.Series.explode)
-                    .reset_index()
-                )
+    def decode_alleles_two_channel(
+        self, allele1: np.ndarray, allele2: np.ndarray
+    ) -> np.ndarray:
+        """Convert two integer allele matrices back into IUPAC-encoded genotypes.
 
-            if write_output:
-                of = outdir / "012.str"
-                df_decoded.insert(0, "sampleID", self._samples)
-                df_decoded.insert(1, "popID", self._populations)
+        This is the inverse of ``encode_alleles_two_channel``: given allele1 and allele2 (each shape (n_samples, n_loci), values in {0,1,2,3} or -1 for missing), reconstruct the original SNP matrix of IUPAC codes: A/C/G/T for homozygotes, ambiguity codes (W,S,M,K,R,Y) for heterozygotes, and "N" for missing.
 
-                df_decoded.to_csv(of, sep="\t", header=False, index=False)
+        Args:
+            allele1: An (n_samples x n_loci) int array of first alleles.
+            allele2: An (n_samples x n_loci) int array of second alleles.
 
-        elif ft.startswith("phylip"):
-            if write_output:
-                of = outdir / "012.phy"
-                header = f"{self.num_inds} {self.num_snps}\n"
-                with open(of, "w") as fout:
-                    fout.write(header)
+        Returns:
+            An (n_samples x n_loci) numpy array of dtype '<U1' with IUPAC codes.
+        """
+        # 1) Build inverse map from numeric pairs → IUPAC letter
+        IUPAC_MAP = self.iupac.get_two_channel_iupac()  # str → (int,int)
+        inv_map: dict[tuple[int, int], str] = {}
+        for base, pair in IUPAC_MAP.items():
+            inv_map[pair] = base
+            # also allow the reversed order for heterozygotes
+            if pair[0] != pair[1]:
+                inv_map[(pair[1], pair[0])] = base
+        # ensure missing maps to "N"
+        inv_map[(-1, -1)] = "N"
 
-                lst_decoded = df_decoded.values.tolist()
+        # 2) Allocate output array of single‐character strings
+        n_samples, n_loci = allele1.shape
+        genotypes = np.full((n_samples, n_loci), "N", dtype="<U1")
 
-                with open(of, "a") as fout:
-                    for sample, row in zip(self._samples, lst_decoded):
-                        seqs = "".join([str(x) for x in row])
-                        fout.write(f"{sample}\t{seqs}\n")
+        # 3) Fill by lookup
+        for i in range(n_samples):
+            for j in range(n_loci):
+                key = (int(allele1[i, j]), int(allele2[i, j]))
+                genotypes[i, j] = inv_map.get(key, "N")
 
-        if write_output:
-            return of
-        return df_decoded.to_numpy()
+        return genotypes
 
     @property
     def genotypes_012(self) -> np.ndarray:
@@ -747,3 +770,46 @@ class GenotypeEncoder:
         Xt = self.inverse_int_iupac(X)
         self.snp_data = Xt
         self.logger.debug(f"Decoded integer-encoded genotypes: {Xt}")
+
+    @property
+    def two_channel_alleles(self) -> Tuple[np.ndarray, np.ndarray]:
+        """Two-channel allele matrices.
+
+        This property encodes the SNP data into two separate matrices representing the two alleles for each sample and locus. Each matrix will have shape (N_samples, N_loci), where each entry is an integer representing one of the two alleles (reference or alternate).
+
+        Warning:
+            This method forces the SNP data to be bi-allelic. Genotypes represented by IUPAC ambiguity codes representing more than two alleles will be set to missing values (-1).
+
+        Returns:
+            Tuple[np.ndarray, np.ndarray]: Two matrices where each row corresponds to a sample and each column to a locus. The first matrix contains the first allele (allele1) and the second matrix contains the second allele (allele2).
+        """
+        alleles = self.encode_alleles_two_channel(self.snp_data)
+        self.logger.debug(f"Alleles (first channel): {alleles[0]}")
+        self.logger.debug(f"Alleles (second channel): {alleles[1]}")
+        return alleles
+
+    @two_channel_alleles.setter
+    def two_channel_alleles(self, value: Tuple[np.ndarray, np.ndarray]) -> None:
+        """Set the two-channel allele matrices.
+
+        This method decodes the two-channel allele matrices back to a 2D list of IUPAC genotypes as ``snp_data``.
+
+        Args:
+            value (Tuple[np.ndarray, np.ndarray]): Two matrices where each row corresponds to a sample and each column to a locus. The first matrix contains the first allele (allele1) and the second matrix contains the second allele (allele2).
+        """
+        if not isinstance(value, tuple):
+            msg = f"Value must be a tuple of two numpy arrays, but got: {type(value)}"
+            self.logger.error(msg)
+            raise TypeError(msg)
+
+        if len(value) != 2:
+            msg = f"Value must be a tuple of two numpy arrays, but got: {len(value)} elements"
+            self.logger.error(msg)
+            raise ValueError(msg)
+
+        allele1 = misc.validate_input_type(value[0], return_type="array")
+        allele2 = misc.validate_input_type(value[1], return_type="array")
+        self.snp_data = self.decode_alleles_two_channel(allele1, allele2)
+        self.logger.debug(
+            f"Decoded two-channel alleles: {self.snp_data} with shape {self.snp_data.shape}"
+        )
