@@ -1,6 +1,7 @@
 import copy
 import itertools
 import logging
+from functools import cached_property
 from pathlib import Path
 from typing import Any, Dict, Generator, List, Literal, Tuple
 
@@ -13,6 +14,7 @@ import snpio.utils.custom_exceptions as exceptions
 from snpio.plotting.plotting import Plotting
 from snpio.read_input.genotype_data_base import BaseGenotypeData
 from snpio.read_input.popmap_file import ReadPopmap
+from snpio.utils.containers import IOConfig, PlotConfig, PopState
 from snpio.utils.logging import LoggerManager
 from snpio.utils.misc import IUPAC
 from snpio.utils.missing_stats import MissingStats
@@ -180,7 +182,7 @@ class GenotypeData(BaseGenotypeData):
             "genepop",
         }
         self._snp_data = None
-        self.from_vcf = False
+        self.from_vcf = getattr(self, "from_vcf", False)
 
         self.logger: logging.Logger | None = None if logger is None else logger
 
@@ -207,15 +209,25 @@ class GenotypeData(BaseGenotypeData):
             "debug": debug,
         }
 
-        self.plot_kwargs = {
-            "plot_format": self.plot_format,
-            "plot_fontsize": self.plot_fontsize,
-            "dpi": self.plot_dpi,
-            "despine": self.plot_despine,
-            "show": self.show_plots,
-            "verbose": self.verbose,
-            "debug": self.debug,
-        }
+        self.plot_config = PlotConfig(
+            plot_format=self.plot_format,
+            plot_fontsize=self.plot_fontsize,
+            dpi=self.plot_dpi,
+            despine=self.plot_despine,
+            show=self.show_plots,
+            verbose=self.verbose,
+            debug=self.debug,
+        )
+
+        self.io_config = IOConfig(
+            prefix=self.prefix,
+            chunk_size=self.chunk_size,
+            force_popmap=self.force_popmap,
+            include_pops=self.include_pops,
+            exclude_pops=self.exclude_pops,
+            verbose=self.verbose,
+            debug=self.debug,
+        )
 
         self.missing_vals = ["N", "-", ".", "?"]
         self.replace_vals = [pd.NA] * len(self.missing_vals)
@@ -248,6 +260,8 @@ class GenotypeData(BaseGenotypeData):
             v: k for k, v in self.iupac_mapping.items()
         }
 
+        self.pop_state = PopState(samples=self._samples)
+
         if not hasattr(self, "iupac"):
             self.iupac = IUPAC(logger=self.logger)
 
@@ -256,6 +270,14 @@ class GenotypeData(BaseGenotypeData):
 
         self.snpio_mqc = SNPioMultiQC
 
+        self.set_alignment(
+            self.snp_data,
+            self.samples,
+            self.sample_indices,
+            self.loci_indices,
+            reset_attributes=False,
+        )
+
     def _iupac_from_gt_tuples(self) -> Dict[Tuple[str, str], str]:
         """Returns the IUPAC code mapping.
 
@@ -334,9 +356,16 @@ class GenotypeData(BaseGenotypeData):
             pm.get_pop_counts(self)
 
         else:
-            msg = "Popmap file not provided but is required. Please provide a valid popmap file."
-            self.logger.error(msg)
-            raise exceptions.PopmapFileNotFoundError(msg)
+            msg = "No popmapfile provided. Skipping population map reading."
+            self.logger.info(msg)
+            self._popmap = dict(zip(self.samples, ["NA"] * self.num_inds))
+            self._popmap_inverse = {"NA": self.samples}
+            self._populations = ["NA"] * self.num_inds
+            self.sample_indices = np.ones(self.num_inds, dtype=bool)
+
+        self.pop_state.populations = self._populations
+        self.pop_state.popmap = self._popmap
+        self.pop_state.popmap_inverse = self._popmap_inverse
 
     def subset_with_popmap(
         self,
@@ -398,23 +427,26 @@ class GenotypeData(BaseGenotypeData):
             self.logger.error(msg)
             raise exceptions.EmptyIterableError(msg)
 
-        # Update the sample indices as a boolean array
-        self.sample_indices = np.isin(self.samples, new_samples)
+        # Create a boolean mask for filtering
+        sample_mask = np.isin(self.samples, new_samples)
 
         # Update samples and populations based on the subset
         self.samples = new_samples
         self._populations = new_populations
 
-        # Ensure the snp_data is filtered by the subset of samples
-        self.snp_data = self.snp_data[self.sample_indices, :]
+        # Filter the snp_data using the mask
+        self.snp_data = self.snp_data[sample_mask, :]
 
         # Update popmap and inverse popmap
         self._popmap = my_popmap.popmap
         self._popmap_inverse = my_popmap.popmap_flipped
 
+        # Reset sample_indices to reflect the new state (all current samples are kept)
+        self.sample_indices = np.ones(self.num_inds, dtype=bool)
+
         # Return indices if requested
         if return_indices:
-            return self.sample_indices
+            return sample_mask  # Return the original mask used for filtering
 
     def write_popmap(self, filename: str) -> None:
         """Write the population map to a file.
@@ -468,8 +500,15 @@ class GenotypeData(BaseGenotypeData):
         """
         self.logger.info(f"Writing VCF file to: {output_filename}")
 
-        from_vcf = getattr(self, "from_vcf", False)
+        from_vcf = self.from_vcf
+
         of = Path(output_filename)
+        parent_pth = of.parent
+        if of.suffix == ".gz":
+            of = Path(of.stem)  # Remove .gz suffix; gets added back later
+
+        of = parent_pth / of
+        of.parent.mkdir(parents=True, exist_ok=True)
 
         try:
             if from_vcf:
@@ -589,7 +628,6 @@ class GenotypeData(BaseGenotypeData):
                                 f.write("\t".join(row) + "\n")
 
                             pbar.update(n)
-
             else:
                 # Raw fallback
                 n_loci = self.snp_data.shape[1]
@@ -640,10 +678,11 @@ class GenotypeData(BaseGenotypeData):
         if not hasattr(self, "_is_bgzipped"):
             self._is_bgzipped = lambda x: False
 
-        # bgzip & tabix
-        bgz = of if self._is_bgzipped(of) else self.bgzip_file(of)
-        self.tabix_index(bgz)
-        self.logger.info(f"Indexed VCF file: {bgz}.tbi")
+        # bgzip the file if not already bgzipped and tabix index it.
+        # NOTE: This adds a .gz suffix to the output file.
+        self.tabix_index(of)
+
+        self.logger.info(f"Indexed VCF file: {of}.tbi")
         self.logger.info("Successfully wrote VCF file!")
 
     def write_phylip(
@@ -704,8 +743,14 @@ class GenotypeData(BaseGenotypeData):
 
         try:
             with open(output_file, "w") as f:
-                n_samples, n_loci = len(samples), len(snp_data[0])
+                n_samples = len(samples)
+                if n_samples == 0:
+                    n_loci = 0
+                else:
+                    n_loci = len(snp_data[0])
+
                 f.write(f"{n_samples} {n_loci}\n")
+
                 for sample, sample_data in zip(samples, snp_data):
                     genotype_str = "".join(str(x) for x in sample_data)
                     f.write(f"{sample}\t{genotype_str}\n")
@@ -773,7 +818,7 @@ class GenotypeData(BaseGenotypeData):
             elif hasattr(self, "marker_names") and self.marker_names:
                 marker_names_list = list(self.marker_names)
             else:
-                marker_names_list = [f"locus_{i}" for i in range(self.num_snps)]
+                marker_names_list = [f"locus_{i+1}" for i in range(self.num_snps)]
 
         if hasattr(self, "allele_start_col"):
             allele_start_col = self.allele_start_col
@@ -1161,9 +1206,9 @@ class GenotypeData(BaseGenotypeData):
     def _get_custom_gt2iupac(self):
         iupac_ambig_dict = {
             "AA": "A",
-            "TT": "T",
             "CC": "C",
             "GG": "G",
+            "TT": "T",
             "AT": "W",
             "TA": "W",
             "AC": "M",
@@ -1253,22 +1298,39 @@ class GenotypeData(BaseGenotypeData):
             self, "vcf_attributes_fn", None
         ):
             hdf = Path(self.vcf_attributes_fn)
-            if hdf.is_file():
-                locus_names = []
-                # Read HDF5 attributes to get locus names
-                with h5py.File(hdf, "r") as h5:
-                    for x, y in zip(h5["chrom"][:], h5["pos"][:]):
-                        if isinstance(x, (str, bytes)):
-                            x = x.decode()
-                        if isinstance(y, (str, bytes)):
-                            y = y.decode()
 
-                        # Ensure x and y are strings
-                        locus_names.append(f"{x}:{y}")
-            else:
-                raise FileNotFoundError(f"HDF5 attribute file not found: {hdf}")
+            if not hdf.is_file():
+                msg = f"HDF5 file with VCF attributes not found: {hdf}"
+                self.logger.error(msg)
+                raise FileNotFoundError(msg)
+
+            locus_names = []
+
+            # Read HDF5 attributes to get locus names
+            with h5py.File(hdf, "r") as h5:
+                # Read all locus metadata from HDF5
+                all_chroms = h5["chrom"][:].astype(str)
+                all_pos = h5["pos"][:].astype(str)
+
+                # Check if a filter has been applied
+                loci_filter = self.loci_indices
+                if loci_filter is not None and loci_filter.size == all_chroms.size:
+                    # Apply the filter to the locus metadata
+                    filtered_chroms = all_chroms[loci_filter]
+                    filtered_pos = all_pos[loci_filter]
+                else:
+                    # No filter or mismatch, use all loci
+                    filtered_chroms = all_chroms
+                    filtered_pos = all_pos
+
+                # Combine to create final locus names
+                locus_names = [
+                    f"{c}:{p}" for c, p in zip(filtered_chroms, filtered_pos)
+                ]
+
+            df.columns = locus_names
         else:
-            locus_names = df.columns.tolist()
+            locus_names = [f"locus_{i+1}" for i in range(df.shape[1])]
 
         # 2. Attach names and compute core proportions
         if len(df.columns) != len(locus_names):
@@ -1294,7 +1356,7 @@ class GenotypeData(BaseGenotypeData):
 
         poploc = poptot = indpop = None
 
-        if use_pops:
+        if self.has_popmap:
             # Boolean mask for missing values
             indpop = df.isna()
 
@@ -1369,14 +1431,17 @@ class GenotypeData(BaseGenotypeData):
         self.logger.error(msg)
         raise NotImplementedError(msg)
 
-    def get_population_indices(self) -> Dict[str, int]:
+    def get_population_indices(self) -> Dict[str, List[int]]:
         """Create a mapping from population IDs to sample indices.
 
         This method creates a dictionary with population IDs as keys and lists of sample indices as values. The sample indices are used to subset the genotype data by population.
 
         Returns:
-            Dict[str, int]: Dictionary with population IDs as keys and lists of sample indices as values.
+            Dict[str, List[int]]: Dictionary with population IDs as keys and lists of sample indices as values.
         """
+        if self.popmap_inverse is None or not self.popmap_inverse:
+            return {"NA": self.samples}
+
         sample_id_to_index = {
             sample_id: idx for idx, sample_id in enumerate(self.samples)
         }
@@ -1390,6 +1455,18 @@ class GenotypeData(BaseGenotypeData):
             pop_indices[pop_id] = indices
         return pop_indices
 
+    @property
+    def locus_names(self) -> list[str]:
+        """Concrete locus names, generating defaults if absent."""
+        if getattr(self, "marker_names", None) is not None:
+            return list(self.marker_names)
+        return [f"locus_{i+1}" for i in range(self.num_snps)]
+
+    @cached_property
+    def valid_mask(self) -> np.ndarray:
+        """Boolean mask [n_samples, n_loci] where True = non-missing genotype."""
+        return ~self.missing_mask
+
     @property
     def inputs(self) -> Dict[str, Any]:
         """Get GenotypeData keyword arguments as a dictionary.
@@ -1408,6 +1485,26 @@ class GenotypeData(BaseGenotypeData):
         """
         self.kwargs = value
 
+    @property
+    def shape(self) -> tuple[int, int]:
+        """(n_samples, n_loci) tuple.
+
+        Returns:
+            tuple[int, int]: Tuple with number of samples (individuals) and number of loci (SNPs).
+        """
+        return (self.num_inds, self.num_snps)
+
+    @cached_property
+    def nbytes(self) -> int:
+        """Approximate RAM footprint of snp_data (bytes).
+
+        Returns:
+            int: Approximate RAM footprint of snp_data (bytes).
+        """
+        arr = self.snp_data
+        # dtype "<U1" is variable-sized in NumPy; use itemsize * size as a lower bound
+        return arr.size * arr.dtype.itemsize
+
     @property
     def num_snps(self) -> int:
         """Number of snps (loci) in the dataset.
@@ -1462,7 +1559,7 @@ class GenotypeData(BaseGenotypeData):
             try:
                 value = int(value)
             except ValueError:
-                f"num_inds must be numeric, but got {type(value)}"
+                msg = f"num_inds must be numeric, but got {type(value)}"
                 self.logger.error(msg)
                 raise TypeError(msg)
         if value < 0:
@@ -1473,104 +1570,49 @@ class GenotypeData(BaseGenotypeData):
         self._num_inds = value
 
     @property
-    def populations(self) -> List[str | int]:
-        """Population IDs as a list of strings or integers.
+    def samples(self) -> list[str]:
+        return self.pop_state.samples
 
-        Returns:
-            List[str | int]: Population IDs.
-        """
-        return self._populations
+    @samples.setter
+    def samples(self, value: list[str]) -> None:
+        self.pop_state.samples = value
+        self._samples = value
+
+    @property
+    def populations(self) -> list[str | int]:
+        return self.pop_state.populations
 
     @populations.setter
-    def populations(self, value: List[str | int]) -> None:
-        """Set the population IDs.
-
-        Args:
-            value (List[str | int]): List of population IDs.
-        """
+    def populations(self, value: list[str | int]) -> None:
+        self.pop_state.populations = value
         self._populations = value
 
     @property
-    def popmap(self) -> Dict[str, str | int]:
-        """Dictionary object with SampleIDs as keys and popIDs as values.
-
-        Returns:
-            Dict[str, str | int]: Dictionary with SampleIDs as keys and popIDs as values
-        """
-        return self._popmap
+    def popmap(self) -> dict[str, str | int] | None:
+        return self.pop_state.popmap
 
     @popmap.setter
-    def popmap(self, value: Dict[str, str | int]) -> None:
-        """Dictionary with SampleIDs as keys and popIDs as values.
-
-        Args:
-            value (Dict[str, str | int]): Dictionary object with SampleIDs as keys and popIDs as values.
-
-        Raises:
-            TypeError: If the value is not a dictionary object.
-            TypeError: If the values are not strings or integers.
-        """
-        if not isinstance(value, dict):
-            msg = f"'popmap' must be a dict object: {type(value)}"
-            self.logger.error(msg)
-            raise TypeError(msg)
-
-        if not all(isinstance(v, (str, int)) for v in value.values()):
-            msg = f"popmap values must be strings or integers"
-            self.logger.error(msg)
-            raise TypeError(msg)
-
+    def popmap(self, value: dict[str, str | int] | None) -> None:
+        self.pop_state.popmap = value
         self._popmap = value
 
     @property
-    def popmap_inverse(self) -> Dict[str, List[str]]:
-        """Inverse popmap dictionary with populationIDs as keys and lists of sampleIDs as values.
-
-        Returns:
-            Dict[str, List[str]]: Inverse dictionary of popmap, where popIDs are keys and lists of sampleIDs are values.
-        """
-        return self._popmap_inverse
+    def popmap_inverse(self) -> dict[str, list[str]] | None:
+        return self.pop_state.popmap_inverse
 
     @popmap_inverse.setter
-    def popmap_inverse(self, value: Dict[str, List[str]]) -> None:
-        """Setter for popmap_inverse. Should have populationIDs as keys and lists of corresponding sampleIDs as values.
-
-        Args:
-            value (Dict[str, List[str]]): Inverse dictionary of popmap, where popIDs are keys and lists of sampleIDs are values.
-
-        Raises:
-            TypeError: If the value is not a dictionary object.
-            TypeError: If the values are not lists of sampleIDs.
-        """
-        if not isinstance(value, dict):
-            msg = f"'popmap_inverse' must be a dict object: {type(value)}"
-            self.logger.error(msg)
-            raise TypeError(msg)
-
-        if not all(isinstance(v, list) for v in value.values()):
-            msg = f"'popmap_inverse' values must be lists of sampleIDs per populationID key, but got: {[type(v) for v in value.values()]}"
-            self.logger.error(msg)
-            raise TypeError(msg)
-
+    def popmap_inverse(self, value: dict[str, list[str]] | None) -> None:
+        self.pop_state.popmap_inverse = value
         self._popmap_inverse = value
 
     @property
-    def samples(self) -> List[str]:
-        """List of sample IDs in input order.
+    def num_pops(self) -> int:
+        """Number of populations in the dataset.
 
         Returns:
-            List[str]: Sample IDs in input order.
+            int: Number of unique populations in the dataset.
         """
-        return self._samples
-
-    @samples.setter
-    def samples(self, value: List[str]) -> None:
-        """Get the sampleIDs as a list of strings.
-
-        Args:
-            value (List[str]): List of sample IDs.
-        """
-        self._samples = value
+        return len(set(self.populations))
 
     @property
     def snpsdict(self) -> Dict[str, List[str]]:
@@ -1615,7 +1657,7 @@ class GenotypeData(BaseGenotypeData):
             self._loci_indices = np.ones(self.num_snps, dtype=bool)
 
         else:
-            if not self._loci_indices.dtype is np.dtype(bool):
+            if self._loci_indices.dtype != bool:
                 msg = f"'loci_indices' must be numpy.dtype 'bool', but got: {self._loci_indices.dtype}"
                 self.logger.error(msg)
                 raise TypeError(msg)
@@ -1624,52 +1666,31 @@ class GenotypeData(BaseGenotypeData):
 
     @loci_indices.setter
     def loci_indices(self, value: np.ndarray | list) -> None:
-        """Column indices for retained loci in filtered alignment.
-
-        Args:
-            value (np.ndarray): Boolean array of loci indices, with True for retained loci and False for excluded loci.
-
-        Raises:
-            TypeError: If the loci_indices attribute is not a numpy.dtype 'bool'.
-        """
         if value is None:
             value = np.ones(self.num_snps, dtype=bool)
         if isinstance(value, list):
             value = np.array(value)
-        if not value.dtype is np.dtype(bool):
-            msg = f"Attempt to set 'loci_indices' to an unexpected np.dtype. Expected 'bool', but got: {value.dtype}"
+        if value.dtype != bool:
+            msg = f"Attempt to set 'loci_indices' to unexpected dtype. Expected bool, got: {value.dtype}"
             self.logger.error(msg)
             raise TypeError(msg)
         self._loci_indices = value
 
     @property
     def sample_indices(self) -> np.ndarray:
-        """Row indices for retained samples in alignment.
-
-        Returns:
-            np.ndarray: Boolean array of sample indices, with True for retained samples and False for excluded samples.
-
-        Raises:
-            TypeError: If the sample_indices attribute is not a numpy.ndarray or list.
-            TypeError: If the sample_indices attribute is not a numpy.dtype 'bool'.
-        """
-        if not isinstance(self._sample_indices, (np.ndarray, list)):
-            msg = f"'sample_indices' set to invalid type. Expected numpy.ndarray or list, but got: {type(self._sample_indices)}"
+        if not isinstance(self._sample_indices, (np.ndarray, list, type(None))):
+            msg = f"'sample_indices' set to invalid type. Expected numpy.ndarray, list, or None, but got: {type(self._sample_indices)}"
             self.logger.error(msg)
             raise TypeError(msg)
-
-        elif isinstance(self._sample_indices, list):
+        if isinstance(self._sample_indices, list):
             self._sample_indices = np.array(self._sample_indices)
-
-        if self._sample_indices is None or not self._sample_indices.size > 0:
-            self._sample_indices = np.ones_like(self.samples, dtype=bool)
-
+        if self._sample_indices is None or self._sample_indices.size == 0:
+            self._sample_indices = np.ones(len(self.samples), dtype=bool)
         else:
-            if not self._sample_indices.dtype is np.dtype(bool):
-                msg = f"'sample_indices' must be numpy.dtype 'bool', but got: {self._sample_indices.dtype}"
+            if self._sample_indices.dtype != bool:
+                msg = f"'sample_indices' must be dtype bool, but got: {self._sample_indices.dtype}"
                 self.logger.error(msg)
                 raise TypeError(msg)
-
         return self._sample_indices
 
     @sample_indices.setter
@@ -1696,16 +1717,20 @@ class GenotypeData(BaseGenotypeData):
         Returns:
             List[str]: List of reference alleles of length num_snps.
         """
+        if isinstance(self._ref, np.ndarray):
+            return self._ref.astype(str).tolist()
         return self._ref
 
-    @ref.setter
-    def ref(self, value: List[str]) -> None:
-        """Setter for list of reference alleles of length num_snps.
+    # In the GenotypeData class
 
-        Args:
-            value (List[str]): List of reference alleles of length num_snps.
-        """
-        self._ref = value
+    @ref.setter
+    def ref(self, value: List[str] | np.ndarray) -> None:
+        """Setter for list of reference alleles of length num_snps."""
+        if isinstance(value, np.ndarray):
+            # .tolist() converts numpy string types to python strings
+            self._ref = value.tolist()
+        else:
+            self._ref = value
 
     @property
     def alt(self) -> List[str]:
@@ -1714,16 +1739,22 @@ class GenotypeData(BaseGenotypeData):
         Returns:
             List[str]: List of alternate alleles of length num_snps.
         """
+        if isinstance(self._alt, np.ndarray):
+            return self._alt.astype(str).tolist()
         return self._alt
 
     @alt.setter
-    def alt(self, value: List[str]) -> None:
-        """Setter for list of alternate alleles of length num_snps.
-
-        Args:
-            value (List[str]): List of alternate alleles of length num_snps.
-        """
-        self._alt = value
+    def alt(self, value: List[List[str]]) -> None:
+        """Setter for list of alternate alleles of length num_snps."""
+        if isinstance(value, list) and value:
+            # Recursively convert any numpy strings to python strings
+            # This handles the nested list structure for multi-allelic sites
+            self._alt = [[str(allele) for allele in inner_list] for inner_list in value]
+        elif isinstance(value, np.ndarray):
+            # Fallback for a simple numpy array
+            self._alt = value.tolist()
+        else:
+            self._alt = value
 
     @property
     def snp_data(self) -> np.ndarray:
@@ -1776,12 +1807,223 @@ class GenotypeData(BaseGenotypeData):
                 value = value.astype("<U1")
             self._snp_data = value
             self._validate_seq_lengths()
+            self._invalidate_caches()
         else:
             msg = f"Attempt to set 'snp_data' to invalid type. Must be a list, numpy.ndarray, pandas.DataFrame, but got: {type(value)}"
             self.logger.debug(value)
             self.logger.error(msg)
             raise TypeError(msg)
 
+    @property
+    def output_dir(self) -> Path:
+        """Root output directory for this dataset.
+
+        Returns:
+            Path: Path to the root output directory.
+        """
+        return Path(f"{self.prefix}_output")
+
+    @property
+    def reports_dir(self) -> Path:
+        """Standardized location for reports (pre/post-filtering aware).
+
+        Returns:
+            Path: Path to the reports directory.
+        """
+        if self.was_filtered:
+            return self.output_dir / "nremover" / "gtdata" / "reports"
+        return self.output_dir / "gtdata" / "reports"
+
+    @property
+    def plot_kwargs(self) -> dict:
+        """Backwards compatibility; convert PlotConfig to the old dict shape.
+
+        Returns:
+            dict: Dictionary of plot configuration parameters.
+        """
+        return self.plot_config.to_dict()
+
+    @property
+    def plots_dir(self) -> Path:
+        """Standardized location for plots (pre/post-filtering aware).
+
+        Returns:
+            Path: Path to the plots directory.
+        """
+        if self.was_filtered:
+            return self.output_dir / "nremover" / "gtdata" / "plots"
+        return self.output_dir / "gtdata" / "plots"
+
+    @cached_property
+    def het_mask(self) -> np.ndarray:
+        """[n_samples, n_loci] True if genotype is heterozygous (IUPAC ambiguity codes).
+
+        Returns:
+            np.ndarray: Boolean mask where True indicates a heterozygous genotype.
+        """
+        ambig = np.array(list("WRMKYS"), dtype="<U1")
+        return np.isin(self.snp_data, ambig)
+
+    @cached_property
+    def per_locus_het_rate(self) -> pd.Series:
+        """Heterozygote proportion per locus (ignores missing).
+
+        Returns:
+            pd.Series: Heterozygote proportion per locus as a pandas Series indexed by locus name.
+        """
+        vm = self.valid_mask
+        hm = self.het_mask
+        denom = vm.sum(axis=0).clip(min=1)
+        vals = (hm & vm).sum(axis=0) / denom
+        return pd.Series(vals, index=self.locus_names, name="het_rate").round(4)
+
+    @cached_property
+    def per_individual_het_rate(self) -> pd.Series:
+        """Heterozygote proportion per individual (ignores missing).
+
+        Returns:
+            pd.Series: Heterozygote proportion per individual as a pandas Series indexed by sample name.
+        """
+        vm = self.valid_mask
+        hm = self.het_mask
+        denom = vm.sum(axis=1).clip(min=1)
+        vals = (hm & vm).sum(axis=1) / denom
+        return pd.Series(vals, index=self.samples, name="het_rate").round(4)
+
+    @cached_property
+    def missing_mask(self) -> np.ndarray:
+        """Boolean mask [n_samples, n_loci] where True indicates a missing genotype.
+
+        Returns:
+            np.ndarray: Boolean mask where True indicates a missing genotype.
+        """
+        # build a mask by multi-compare against the small missing set
+        arr = self.snp_data
+        misses = np.isin(arr, np.array(self.missing_vals, dtype=arr.dtype))
+        return misses
+
+    @property
+    def missing_rate(self) -> float:
+        """Overall missing proportion in the alignment.
+
+        Returns:
+            float: Overall missing proportion in the alignment.
+        """
+        mm = self.missing_mask
+        if mm.size == 0:
+            return 0.0
+        return float(mm.mean())
+
+    @property
+    def per_individual_missing(self) -> pd.Series:
+        """Missing proportion per sample as a pandas Series indexed by sample name.
+
+        Returns:
+            pd.Series: Missing proportion per sample as a pandas Series indexed by sample name.
+        """
+        mm = self.missing_mask
+        vals = mm.mean(axis=1)
+        return pd.Series(vals, index=self.samples, name="missing_prop").round(4)
+
+    @property
+    def per_locus_missing(self) -> pd.Series:
+        """Missing proportion per locus as a pandas Series; uses marker names if present.
+
+        Returns:
+            pd.Series: Missing proportion per locus as a pandas Series indexed by locus name.
+        """
+        mm = self.missing_mask
+        cols = (
+            self.marker_names
+            if getattr(self, "marker_names", None) is not None
+            else [f"locus_{i+1}" for i in range(mm.shape[1])]
+        )
+        vals = mm.mean(axis=0)
+        return pd.Series(vals, index=cols, name="missing_prop").round(4)
+
+    @property
+    def sample_index_map(self) -> Dict[str, int]:
+        """Map sample ID -> row index (useful for subsetting).
+
+        Returns:
+            Dict[str, int]: Mapping of sample IDs to their corresponding row indices.
+        """
+        return {s: i for i, s in enumerate(self.samples)}
+
+    @property
+    def pop_sizes(self) -> Dict[str, int]:
+        """Population -> sample count.
+
+        Returns:
+            Dict[str, int]: Mapping of population names to sample counts.
+        """
+        counts: Dict[str, int] = {}
+        for p in self.populations:
+            counts[p] = counts.get(p, 0) + 1
+        return counts
+
+    @property
+    def pop_to_indices(self) -> Dict[str, list[int]]:
+        """Population -> list of sample indices (built from current popmap_inverse).
+
+        Returns:
+            Dict[str, list[int]]: Mapping of population names to lists of sample indices.
+        """
+        out: Dict[str, list[int]] = {}
+        if self.popmap_inverse is None:
+            return out
+        idx_map = self.sample_index_map
+        for pop, ids in self.popmap_inverse.items():
+            out[pop] = [idx_map[s] for s in ids if s in idx_map]
+        return out
+
+    @property
+    def is_empty(self) -> bool:
+        """True if there are zero samples or loci.
+
+        Returns:
+            bool: True if there are zero samples or loci.
+        """
+        return self.num_inds == 0 or self.num_snps == 0
+
+    @property
+    def has_popmap(self) -> bool:
+        """True if population information is present.
+
+        Returns:
+            bool: True if population information is present.
+        """
+        return self.popmapfile is not None
+
+    @cached_property
+    def is_missing_locus(self) -> np.ndarray:
+        """[n_loci] True if an entire locus is missing across all samples."""
+        return (~self.valid_mask).all(axis=0)
+
+    @cached_property
+    def observed_iupac_per_locus(self) -> list[set[str]]:
+        """Observed IUPAC codes per locus (excluding missing)."""
+        arr = self.snp_data
+        vm = self.valid_mask
+        return [set(arr[vm[:, j], j]) for j in range(arr.shape[1])]
+
+    @cached_property
+    def biallelic_mask(self) -> np.ndarray:
+        """[n_loci] True where locus appears biallelic (A/C/G/T plus heterozygotes of those two)."""
+        obs = self.observed_iupac_per_locus
+
+        # Extract unambiguous nucleotides present
+        def nucs(s):
+            return {x for x in s if x in {"A", "C", "G", "T"}}
+
+        return np.array([len(nucs(s)) == 2 for s in obs], dtype=bool)
+
+    @property
+    def has_multiallelic(self) -> bool:
+        """True if any locus shows >2 unambiguous nucleotides."""
+        obs = self.observed_iupac_per_locus
+        return any(len({x for x in s if x in {"A", "C", "G", "T"}}) > 2 for s in obs)
+
     def _validate_seq_lengths(self) -> None:
         """Ensure that all SNP data rows have the same length.
 
@@ -1798,6 +2040,24 @@ class GenotypeData(BaseGenotypeData):
                     self.logger.error(msg)
                     raise exceptions.SequenceLengthError(self.samples[i])
 
+    def _invalidate_caches(self) -> None:
+        """Clear cached_property results after data changes."""
+        for attr in (
+            "valid_mask",
+            "is_missing_locus",
+            "observed_iupac_per_locus",
+            "biallelic_mask",
+            "per_locus_missing",
+            "per_individual_missing",
+            "het_mask",
+            "per_locus_het_rate",
+            "per_individual_het_rate",
+            "maf_per_locus",
+            "nbytes",
+        ):
+            if attr in self.__dict__:  # cached_property stores value here
+                self.__dict__.pop(attr, None)
+
     def set_alignment(
         self,
         snp_data: np.ndarray,
@@ -1840,16 +2100,12 @@ class GenotypeData(BaseGenotypeData):
             for s, p in self.popmap.items():
                 self.popmap_inverse.setdefault(p, []).append(s)
 
-        ref, alt, alt2 = self.get_ref_alt_alleles(self.snp_data)
-        self.ref = ref
-        self.alt = alt
-        self.alt = list(alt)
-        if alt2:
-            self.alt = [a if isinstance(a, list) else [a] for a in self.alt]
-            for i, extra in enumerate(alt2):
-                self.alt[i].extend(
-                    extra.tolist() if isinstance(extra, np.ndarray) else extra
-                )
+        # Determine REF and ALT from SNP data matrix if not loading from a VCF.
+        # If loading VCF, self._ref and self._alt are set in get_vcf_attributes.
+        if not self.from_vcf:
+            ref_alleles, alt_alleles = self.get_ref_alt_alleles(self.snp_data)
+            self.ref = ref_alleles
+            self.alt = alt_alleles
 
         if reset_attributes and self.filetype.startswith("vcf"):
             self.vcf_attributes_fn = self.update_vcf_attributes(
@@ -1859,6 +2115,8 @@ class GenotypeData(BaseGenotypeData):
                 samples=self.samples,
             )
 
+        self._invalidate_caches()
+
     def __len__(self) -> int:
         """Return the number of individuals in the dataset.
 
@@ -1891,7 +2149,7 @@ class GenotypeData(BaseGenotypeData):
 
     def __str__(self) -> str:
         """Return a string representation of the GenotypeData object."""
-        return f"GenotypeData: {self.num_snps} SNPs, {self.num_inds} individuals"
+        return f"GenotypeData: {self.num_snps} SNPs, {self.num_inds} individuals, from file: {self.filename} and popmapfile: {self.popmapfile}"
 
     def __repr__(self) -> str:
         """Return a detailed string representation of the GenotypeData object."""
