@@ -1,5 +1,7 @@
 import logging
+import os
 import re
+import shutil
 import tempfile
 import textwrap
 from collections import Counter
@@ -66,47 +68,51 @@ class BaseGenotypeData:
         """Method to load data from file based on filetype"""
         raise NotImplementedError("Subclasses should implement this method")
 
+    # in BaseGenotypeData class
     def get_ref_alt_alleles(
-        self,
-        data: np.ndarray,
-    ) -> Tuple[np.ndarray, np.ndarray, List[np.ndarray]]:
-        """Determine ref/alt alleles for each locus (column) of IUPAC-encoded genotypes.
+        self, data: np.ndarray
+    ) -> Tuple[np.ndarray, List[List[str]]]:
+        """Determine ref/alt alleles for each locus from IUPAC-encoded genotypes.
 
         Args:
-            data: shape (n_samples, n_variants), dtype=object of single-char IUPAC codes.
-
-        Notes:
-            - The most common allele is chosen as the reference (REF).
-            - The second most common allele is chosen as the alternate (ALT).
-            - If there are more than two alleles, any further alleles are stored in a list.
-            - If there are ties in counts, the one with the fewest heterozygotes is chosen.
-            - If still tied, one is chosen at random.
+            data (np.ndarray): (n_variants x n_samples) matrix of IUPAC
 
         Returns:
-            Tuple of
-                - ref_alleles: shape (n_variants,) most‐common allele per locus
-                - alt_alleles: shape (n_variants,) second‐most‐common allele or None
-                - extra_alts: list of arrays of any further alleles per locus
+            tuple: (ref_alleles, alt_lists)
+                - ref_alleles: np.ndarray of REF alleles for each locus
+                - alt_lists: List of ALT alleles for each locus
         """
-        # Handle edge cases for loci with no valid alleles or only one allele
         if data.size == 0:
-            return np.array([]), np.array([]), []
+            return np.array([], dtype="<U1"), []
 
+        VALID_BASES = {"A", "C", "G", "T"}
         ref_alleles = []
-        alt_alleles = []
-        extra_alts = []
+        alt_alleles_list = []
 
         for locus in data.T:
             unique, counts = np.unique(locus, return_counts=True)
-            sorted_alleles = sorted(zip(unique, counts), key=lambda x: (-x[1], x[0]))
+            valid_allele_counts = [
+                (allele, count)
+                for allele, count in zip(unique, counts)
+                if allele.upper() in VALID_BASES
+            ]
+
+            if not valid_allele_counts:
+                ref_alleles.append("N")
+                alt_alleles_list.append(["."])  # Use placeholder
+                continue
+
+            sorted_alleles = sorted(valid_allele_counts, key=lambda x: (-x[1], x[0]))
 
             ref_alleles.append(sorted_alleles[0][0])
-            alt_alleles.append(
-                sorted_alleles[1][0] if len(sorted_alleles) > 1 else None
-            )
-            extra_alts.append([allele for allele, _ in sorted_alleles[2:]])
 
-        return np.array(ref_alleles), np.array(alt_alleles), extra_alts
+            alts = [allele for allele, _ in sorted_alleles[1:]]
+            if not alts:
+                alt_alleles_list.append(["."])
+            else:
+                alt_alleles_list.append(alts)
+
+        return np.array(ref_alleles, dtype="<U1"), alt_alleles_list
 
     def refs_alts_from_snp_data(
         self, snp_matrix: np.ndarray
@@ -204,23 +210,32 @@ class BaseGenotypeData:
         return out
 
     def build_vcf_header(self) -> str:
-        """Dynamically builds the VCF header based on the loaded sample IDs.
+        """Dynamically builds the VCF header using the stored header object."""
+        # Use the stored header from the original VCF file.
+        if hasattr(self, "vcf_header") and self.vcf_header:
+            header_str = str(self.vcf_header)
+            header_lines = header_str.strip().split("\n")
 
-        Returns:
-            str: The VCF header.
-        """
-        sample_headers = "\t".join(self.samples)
+            # Find and replace the #CHROM line to match the current samples
+            for i, line in enumerate(header_lines):
+                if line.startswith("#CHROM"):
+                    fixed_fields = line.split("\t")[:9]
+                    new_chrom_line = "\t".join(fixed_fields + list(self.samples))
+                    header_lines[i] = new_chrom_line
+                    break
 
-        vcf_header = textwrap.dedent(
-            f"""\
-            ##fileformat=VCFv4.2
-            ##source=SNPio
-            ##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
-            ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
-            #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t{sample_headers}\n"""
-        )
-
-        return vcf_header
+            return "\n".join(header_lines) + "\n"
+        else:
+            # Fallback for data not originating from a VCF file
+            sample_headers = "\t".join(self.samples)
+            return textwrap.dedent(
+                f"""\
+                ##fileformat=VCFv4.2
+                ##source=SNPio
+                ##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
+                ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
+                #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t{sample_headers}\n"""
+            )
 
     def replace_alleles(self, row: List[str], ref: str, alts: List[str]) -> List[str]:
         """Replace the alleles in the VCF row with the corresponding VCF genotype codes.
@@ -299,57 +314,112 @@ class BaseGenotypeData:
 
         return row
 
-    def bgzip_file(self, filepath: Path) -> Path:
-        """BGZips a VCF file using pysam's BGZFile, avoiding overwriting the input file.
+    def _looks_like_bgzf(self, path: Path) -> bool:
+        """Best-effort check: can we open and read with BGZFile?
 
         Args:
-            filepath (Path): The path to the input VCF file (uncompressed).
+            path (Path): Path to the file to check.
 
         Returns:
-            Path: The path to the bgzipped output VCF file with a .vcf.gz suffix.
+            bool: True if the file appears to be BGZF-compressed, False otherwise.
         """
+        try:
+            with pysam.BGZFile(str(path), "rb") as f:
+                _ = f.read(1)
+            return True
+        except Exception:
+            return False
+
+    def _ensure_file_ext(self, filepath: Path) -> Path:
+        """Ensures the output file has a .vcf.gz extension.
+
+        Args:
+            filepath (Path): The original file path.
+
+        Returns:
+            Path: The path to the BGZipped output file.
+        """
+        tokens = filepath.name.split(".")
+        while tokens and tokens[-1] in {"vcf", "gz", "nremover"}:
+            tokens.pop()
+        base = ".".join(tokens) if tokens else filepath.stem
+        return filepath.parent / f"{base}.vcf"
+
+    def bgzip_file(self, filepath: Path) -> Path:
+        """BGZips a VCF file using pysam's BGZFile, preserving parent directories.
+
+        Args:
+            filepath (Path): Path to the input VCF file.
+
+        Returns:
+            Path: Path to the BGZipped output file.
+
+        Notes:
+            - *.vcf      -> write sibling *.vcf.gz (as before)
+            - *.vcf.gz   -> if not BGZF, recompress *in place* atomically
+                        -> if already BGZF, return as-is
+        """
+        filepath = Path(filepath) if not isinstance(filepath, Path) else filepath
+
         if not filepath.exists():
             msg = f"Input VCF file not found: {filepath}"
             self.logger.error(msg)
             raise FileNotFoundError(msg)
 
-        # Skip if already has correct suffix
-        if filepath.name.endswith(".nremover.vcf.gz"):
-            self.logger.warning(
-                f"File already appears bgzipped with expected name: {filepath}"
+        suffixes = filepath.suffixes
+        is_vcfgz_name = len(suffixes) >= 2 and suffixes[-2:] == [".vcf", ".gz"]
+        is_gz_name = suffixes and suffixes[-1] == ".gz"
+
+        # Case A: Caller passed *.vcf.gz
+        if is_vcfgz_name or is_gz_name and filepath.name.endswith(".vcf.gz"):
+            if self._looks_like_bgzf(filepath):
+                # Truly already BGZF-compressed
+                return self._ensure_file_ext(filepath)
+
+            # Not BGZF: recompress in place using a temp file + atomic replace
+            self.logger.info(f"Recompressing to BGZF in place: {filepath}")
+
+            tmpdir = filepath.parent
+            tmp_out = Path(
+                tempfile.mkstemp(dir=tmpdir, prefix=filepath.stem + ".", suffix=".bgz")[
+                    1
+                ]
             )
-            return filepath
+            try:
+                with (
+                    open(filepath, "rb") as f_in,
+                    pysam.BGZFile(str(tmp_out), "wb") as f_out,
+                ):
+                    shutil.copyfileobj(f_in, f_out, length=1024 * 1024)
+                os.replace(tmp_out, filepath)  # atomic swap
+            except Exception:
+                # Clean up tmp on failure
+                try:
+                    tmp_out.unlink(missing_ok=True)
+                finally:
+                    raise
+            return self._ensure_file_ext(filepath)
 
-        name_parts = filepath.name.split(".")
-        while name_parts and name_parts[-1] in {"vcf", "gz", "nremover"}:
-            name_parts.pop()
-        base = ".".join(name_parts)
-        bgzipped_path = f"{base}.vcf.gz"
-        bgzipped_path = Path(bgzipped_path)
+        # Case B: Plain *.vcf -> produce sibling <base>.vcf.gz
+        bgzipped_path = self._ensure_file_ext(filepath)
 
-        # Safety check: don't overwrite input file
         if bgzipped_path.resolve() == filepath.resolve():
-            msg = f"Output path {bgzipped_path} would overwrite input file!"
-            self.logger.error(msg)
-            raise ValueError(msg)
+            if not bgzipped_path.name.endswith(".gz"):
+                bgzipped_path = bgzipped_path.with_suffix(bgzipped_path.suffix + ".gz")
+            else:
+                msg = f"Output path {bgzipped_path} would overwrite input file."
+                self.logger.error(msg)
+                raise IOError(msg)
 
-        try:
-            with (
-                open(filepath, "rb") as f_in,
-                pysam.BGZFile(str(bgzipped_path), "wb") as f_out,
-            ):
-                chunk_size = 1024 * 1024  # 1MB
-                while True:
-                    chunk = f_in.read(chunk_size)
-                    if not chunk:
-                        break
-                    f_out.write(chunk)
+        bgzipped_path.parent.mkdir(parents=True, exist_ok=True)
 
-            return bgzipped_path
+        with (
+            open(filepath, "rb") as f_in,
+            pysam.BGZFile(str(bgzipped_path), "wb") as f_out,
+        ):
+            shutil.copyfileobj(f_in, f_out, length=1024 * 1024)
 
-        except Exception as e:
-            self.logger.error(f"Error bgzipping file {filepath}: {e}")
-            raise e
+        return bgzipped_path
 
     def _is_sorted(self, filepath: Path) -> bool:
         """Checks if the VCF file is sorted alphanumerically by chromosome and position.
@@ -420,15 +490,19 @@ class BaseGenotypeData:
                 for line in header_lines:
                     temp_vcf.write(line + "\n")
                 for record in sorted_data_lines:
-                    temp_vcf.write(str(record))  # record already ends in newline
+                    temp_vcf.write(str(record))  # record already has newline
             temp_vcf_path = Path(temp_vcf.name)
 
             # Compress and index
-            pysam.tabix_compress(str(temp_vcf_path), str(sorted_path), force=True)
-            pysam.tabix_index(str(sorted_path), preset="vcf", force=True)
+            pysam.tabix_compress(
+                str(temp_vcf_path), str(self._ensure_file_ext(sorted_path)), force=True
+            )
+            pysam.tabix_index(
+                str(self._ensure_file_ext(sorted_path)), preset="vcf", force=True
+            )
 
             temp_vcf_path.unlink()  # Clean up temp file
-            return sorted_path
+            return self._ensure_file_ext(sorted_path)
 
         except Exception as e:
             self.logger.error(f"Error sorting VCF file: {e}")
@@ -453,6 +527,9 @@ class BaseGenotypeData:
 
         Args:
             filepath (Path): The path to the bgzipped VCF file.
+
+        Raises:
+            Exception: If there is an error during indexing.
         """
         try:
             pysam.tabix_index(str(filepath), preset="vcf", force=True)
