@@ -1,7 +1,7 @@
 import logging
 from collections.abc import Mapping, Sequence
 from itertools import permutations
-from typing import Any, Dict, List
+from typing import Any, Dict, List, Tuple
 
 import numpy as np
 import pandas as pd
@@ -77,6 +77,31 @@ class IUPAC:
             frozenset(["G", "T"]): "K",
         }
 
+    def get_two_channel_iupac(self) -> Dict[str, Tuple[int, int]]:
+        """Get a two-channel IUPAC dictionary to map genotypes to tuples of integers (0-3).
+
+        This method returns a dictionary mapping IUPAC codes to tuples of integers representing the two channels.
+
+        Returns:
+            Dict[str, Tuple[int, int]]: A dictionary mapping IUPAC codes to tuples of integers.
+        """
+        return {
+            "A": (0, 0),  # A/A
+            "C": (1, 1),  # C/C
+            "G": (2, 2),  # G/G
+            "T": (3, 3),  # T/T
+            "W": (0, 3),  # A/T
+            "S": (1, 2),  # C/G
+            "M": (0, 1),  # A/C
+            "K": (2, 3),  # T/G
+            "R": (0, 2),  # A/G
+            "Y": (1, 3),  # C/T
+            "N": (-1, -1),  # Missing
+            "-": (-1, -1),  # Missing
+            ".": (-1, -1),  # Missing
+            "?": (-1, -1),  # Missing
+        }
+
     def get_phased_encoding(self) -> Dict[str, str]:
         """Get a dictionary of phased encoding for IUPAC codes.
 
@@ -87,9 +112,9 @@ class IUPAC:
         """
         phased_encoding = {
             "A": "A/A",
-            "T": "T/T",
             "C": "C/C",
             "G": "G/G",
+            "T": "T/T",
             "N": "N/N",
             ".": "N/N",
             "?": "N/N",
@@ -216,22 +241,22 @@ class IUPAC:
             >>> # Outputs: {'A': 'A', 'C': 'C', 'G': 'G', 'T': 'T', 'M': 'AC', 'R': 'AG', 'W': 'AT', 'S': 'CG', 'Y': 'CT', 'K': 'GT', 'V': 'ACG', 'H': 'ACT', 'D': 'AGT', 'B': 'CGT', 'X': 'GATC', 'N': 'GATC'}
         """
         return {
-            "A": "A",
-            "C": "C",
-            "G": "G",
-            "T": "T",
-            "M": "AC",
-            "R": "AG",
-            "W": "AT",
-            "S": "CG",
-            "Y": "CT",
-            "K": "GT",
-            "V": "ACG",
-            "H": "ACT",
-            "D": "AGT",
-            "B": "CGT",
-            "X": "GATC",
-            "N": "GATC",
+            "A": ("A", "A"),
+            "C": ("C", "C"),
+            "G": ("G", "G"),
+            "T": ("T", "T"),
+            "M": ("A", "C"),
+            "R": ("A", "G"),
+            "W": ("A", "T"),
+            "S": ("C", "G"),
+            "Y": ("C", "T"),
+            "K": ("G", "T"),
+            "V": ("A", "C", "G"),
+            "H": ("A", "C", "T"),
+            "D": ("A", "G", "T"),
+            "B": ("C", "G", "T"),
+            "X": ("G", "A", "T", "C"),
+            "N": ("G", "A", "T", "C"),
         }
 
     def get_gt2iupac(self) -> Dict[str, str]:
@@ -302,13 +327,13 @@ class IUPAC:
         Example:
             >>> iupac_data = IUPAC()
             >>> print(iupac_data.int_iupac_dict)
-            >>> # Outputs: {'A': 0, 'T': 1, 'G': 2, 'C': 3, 'W': 4, 'R': 5, 'M': 6, 'K': 7, 'Y': 8, 'S': 9, 'N': -9}
+            >>> # Outputs: {'A': 0, 'C': 1, 'G': 2, 'T': 3, 'W': 4, 'R': 5, 'M': 6, 'K': 7, 'Y': 8, 'S': 9, 'N': -9}
         """
         return {
             "A": 0,
-            "T": 1,
+            "C": 1,
             "G": 2,
-            "C": 3,
+            "T": 3,
             "W": 4,
             "R": 5,
             "M": 6,
@@ -330,15 +355,15 @@ class IUPAC:
 
         return {
             "A": [1.0, 0.0, 0.0, 0.0],
-            "T": [0.0, 1.0, 0.0, 0.0],
+            "C": [0.0, 1.0, 0.0, 0.0],
             "G": [0.0, 0.0, 1.0, 0.0],
-            "C": [0.0, 0.0, 0.0, 1.0],
-            "W": [val, val, 0.0, 0.0],
+            "T": [0.0, 0.0, 0.0, 1.0],
+            "W": [val, 0.0, 0.0, val],
             "R": [val, 0.0, val, 0.0],
-            "M": [val, 0.0, 0.0, val],
-            "K": [0.0, val, val, 0.0],
+            "M": [val, val, 0.0, 0.0],
+            "K": [0.0, 0.0, val, val],
             "Y": [0.0, val, 0.0, val],
-            "S": [0.0, 0.0, val, val],
+            "S": [0.0, val, val, 0.0],
             "N": [0.0, 0.0, 0.0, 0.0],
         }
 
